//////////////////////license & copyright header///////////////////////
//                                                                   //
//              Java Cannery - jar management utility.               //
//                Copyright (c) 1998 by Kevin Kelley                 //
//                                                                   //
// This program is free software; you can redistribute it and/or     //
// modify it under the terms of the GNU General Public License as    //
// published by the Free Software Foundation; either version 2 of    //
// the License, or (at your option) any later version.               //
//                                                                   //
// This program is distributed in the hope that it will be useful,   //
// but WITHOUT ANY WARRANTY; without even the implied warranty of    //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     //
// GNU General Public License for more details.                      //
//                                                                   //
// You should have received a copy of the GNU General Public License //
// along with this program in the file 'gpl.html'; if not, write to  //
// the Free Software Foundation, Inc., 59 Temple Place - Suite 330,  //
// Boston, MA 02111-1307, USA, or contact the author:                //
//                                                                   //
//                       Kevin Kelley  <kelley@iguana.ruralnet.net>  //
//                                                                   //
////////////////////end license & copyright header/////////////////////
package starlight.java;
import java.io.*;

class AttributeInfo {  
    static String ATTR_SourceFile    = "SourceFile";
    static String ATTR_ConstantValue = "ConstantValue";
    static String ATTR_CodeAttribute = "Code";
    static String ATTR_LineNumTable  = "LineNumberTable";
    static String ATTR_LocalVarTable = "LocalVariables";
    static String ATTR_Exceptions    = "Exceptions";
    
    JavaClass classInfo;
    int nameIndex;
    int attrLength;
    byte[] data;

    static AttributeInfo createFrom(DataInputStream in, JavaClass cls) {
        AttributeInfo c = null;
        try {
            int nameIndex  = in.readUnsignedShort();
            int attrLength = in.readInt();
            byte[] data = new byte[attrLength];
            in.read(data);

            String name = (String)cls.constantAtIndex(nameIndex).getValue();

                 if (name.equals(ATTR_CodeAttribute)) { c = new CodeAttribute              (); }
            else if (name.equals(ATTR_LineNumTable )) { c = new LineNumberTableAttribute   (); }
            else if (name.equals(ATTR_LocalVarTable)) { c = new LocalVariableTableAttribute(); }
            else if (name.equals(ATTR_ConstantValue)) { c = new ConstantValueAttribute     (); }
            else if (name.equals(ATTR_SourceFile   )) { c = new SourceFileAttribute        (); } 
            else if (name.equals(ATTR_Exceptions   )) { c = new ThrowsAttribute            (); } 
            else                                      { c = new UnrecognizedAttribute      (); }
            
            if (c != null) {
                c.classInfo = cls;
                c.nameIndex = nameIndex;
                c.attrLength = attrLength;
                c.data = data;
            }
        } catch (Exception e) { throw new Error(e.toString()); }
        return c;
    }
    public void writeTo(DataOutputStream out) {
        try {
            out.writeShort(nameIndex);
            out.writeInt(attrLength);
            out.write(data, 0, data.length);
        } catch (Exception e) { throw new Error(e.toString()); }
    }
    public String toString() { return "??"; }
}

class UnrecognizedAttribute extends AttributeInfo {
    public String toString() { return "unrecognized"; } 
}

class SourceFileAttribute extends AttributeInfo {  

    private String sourceFile;
    public String getSourceFile() {  // no path; just  ClassName.java
        if (sourceFile == null)
            createFromData();
        return sourceFile;
    }
    void createFromData() {
        try {
            DataInputStream in = new DataInputStream(new ByteArrayInputStream(data));
            int sourcefileIndex = in.readUnsignedShort();
            sourceFile = (String)classInfo.constantAtIndex(sourcefileIndex).getValue();
        } 
        catch (Exception e) { throw new Error(e.toString()); }
    }
    public String toString() { return getSourceFile(); }
} 


class ConstantValueAttribute extends AttributeInfo {  

    private ConstantInfo constant;
    public Object getConstantValue() {   // Integer, Long, Float, Double, String
        if (constant == null)
            createFromData();
        return constant.getValue();
    }
    void createFromData() {
        try {
            DataInputStream in = new DataInputStream(new ByteArrayInputStream(data));
            int constantIndex = in.readUnsignedShort();
            constant = classInfo.constantAtIndex(constantIndex);
        } 
        catch (Exception e) { throw new Error(e.toString()); }
    }
    public String toString() { return getConstantValue().toString(); }
} 


class CodeAttribute extends AttributeInfo {  
    boolean loaded = false;
    
    int      maxStack;  
    int      maxLocals;  
    int      codeLength;  
    byte     code[];                // multiple-of-four boundary
    int      exceptTableLen;  
    ExceptionTableEntry  exceptionTable[];  
    int       attrCount;  
    AttributeInfo attributes[];     // "LineNumberTable","LocalVariableTable"
                                    //both of which contain debugging information.  

    public int getStackSize() {
        if (!loaded) createFromData(); 
        return maxStack; 
    }
    public int getNumLocalVariables() {
        if (!loaded) createFromData(); 
        return maxLocals; 
    }
    public DataInputStream getCodeStream() {
        if (!loaded) createFromData();
        return new DataInputStream(new ByteArrayInputStream(code));
    }
    public DisAssembler getCodeDisAssembler() {
        if (!loaded) createFromData();
        return new DisAssembler(code);
    }
    public int getNumExceptionHandlers() {
        if (!loaded) createFromData(); 
        return exceptTableLen; 
    }
    public ExceptionTableEntry getExceptionHandler(int i) {
        if (!loaded) createFromData();
        return exceptionTable[i];
    }
    public int getNumAttributes() {
        if (!loaded) createFromData(); 
        return attrCount; 
    }
    public AttributeInfo getAttribute(int i) {
        if (!loaded) createFromData();
        return attributes[i];
    }

    void createFromData() {
        try {
            DataInputStream in = new DataInputStream(new ByteArrayInputStream(data));
            maxStack   = in.readUnsignedShort();
            maxLocals  = in.readUnsignedShort();
            codeLength = in.readInt();
            code = new byte[codeLength];
            in.read(code);
            
            exceptTableLen = in.readUnsignedShort();
            exceptionTable = new ExceptionTableEntry[exceptTableLen];
            for (int i=0; i<exceptTableLen; i++) 
                exceptionTable[i] = ExceptionTableEntry.createFrom(in, classInfo);
            
            attrCount  = in.readUnsignedShort();
            attributes = new AttributeInfo[attrCount];
            for (int i=0; i<attrCount; i++) 
                attributes[i] = AttributeInfo.createFrom(in, classInfo);
            loaded = true;
        } 
        catch (Exception e) { throw new Error(e.toString()); }
    }

    public String hexDumpCode() {
        if (!loaded) createFromData();
        String str = "";
        for (int i=0; i<codeLength; i++) {
            if (i%16 == 0)      str += "\ncode:";
            if (i%16 == 8)      str += " - ";
            str += " ";
            String s = Long.toString(0xff & (long)code[i], 16);
            if (s.length() == 1) str += "0" + s;
            else                 str += s;
        } return str;
    }
    public static String addressString(int addr, int base, int length) {
        String s = Long.toString(addr, base);
        return ("00000000".substring(0, length - s.length())) + s;
    }
    public String toString() {
        if (!loaded) createFromData();

        String str = "  stack size " + maxStack
                   + ";  " + maxLocals + " locals";

        DisAssembler dsm = new DisAssembler(code);
        while (!dsm.eoi())
            str += "\n    " + addressString(dsm.at(), 16, 4)
                + "  " + dsm.readInstruction();

        str += "\n Exception table: ";
        for (int i=0; i<exceptTableLen; i++) 
            str += "\n    " + exceptionTable[i];
        str += "\n code attributes: ";
        for (int i=0; i<attrCount; i++) 
            str += attributes[i];
        return str;
    }
}  


class ThrowsAttribute extends AttributeInfo {  
    boolean loaded = false;
    
    int tableLength;  
    int exceptionIndexes[];  
    
    public int getNumEntries() {
        if (!loaded) createFromData(); 
        return tableLength; 
    }
    // this should return the name of the i'th exception declared 
    // in a method's throws clause.
    public String elementAt(int i) {
        if (!loaded) createFromData();
        return classInfo.constantAtIndex(exceptionIndexes[i]).toString();
    }

    void createFromData() {
        try {
            DataInputStream in = new DataInputStream(new ByteArrayInputStream(data));
            tableLength      = in.readUnsignedShort();
            exceptionIndexes = new int[tableLength];
            for (int i=0; i<tableLength; i++) 
                exceptionIndexes[i] = in.readUnsignedShort();
            loaded = true;
        } 
        catch (Exception e) { throw new Error(e.toString()); }
    }
    public String toString() { 
        if (!loaded) createFromData();

        String str = "";
        if (getNumEntries() > 0) { 
            str += "throws ";
            for (int i=0; i<getNumEntries(); i++) {
                str += elementAt(i);
                if (i < getNumEntries()-1)
                    str += ", ";
            }
        }
        return str;
    }
}


class LineNumberTableAttribute extends AttributeInfo {  
    boolean loaded = false;
    
    int                  tableLength;  
    LineNumberTableEntry lineNumberTable[];  
    
    public int getNumEntries() {
        if (!loaded) createFromData(); 
        return tableLength; 
    }
    public LineNumberTableEntry elementAt(int i) {
        if (!loaded) createFromData();
        return lineNumberTable[i];
    }

    void createFromData() {
        try {
            DataInputStream in = new DataInputStream(new ByteArrayInputStream(data));
            tableLength     = in.readUnsignedShort();
            lineNumberTable = new LineNumberTableEntry[tableLength];
            for (int i=0; i<tableLength; i++) 
                lineNumberTable[i] = LineNumberTableEntry.createFrom(in, classInfo);
            loaded = true;
        } 
        catch (Exception e) { throw new Error(e.toString()); }
    }
    public String toString() { 
        if (!loaded) createFromData();

        String str = "";
        for (int i=0; i<tableLength; i++) 
            str += "\n    " + lineNumberTable[i];
        return str;
    }
}


class LocalVariableTableAttribute extends AttributeInfo {  
    boolean loaded = false;
    
    int                     tableLength;  
    LocalVariableTableEntry localVariableTable[];  

    public int getNumEntries() {
        if (!loaded) createFromData(); 
        return tableLength; 
    }
    public LocalVariableTableEntry elementAt(int i) {
        if (!loaded) createFromData();
        return localVariableTable[i];
    }
    
    void createFromData() {
        try {
            DataInputStream in = new DataInputStream(new ByteArrayInputStream(data));
            tableLength        = in.readUnsignedShort();
            localVariableTable = new LocalVariableTableEntry[tableLength];
            for (int i=0; i<tableLength; i++) 
                localVariableTable[i] = LocalVariableTableEntry.createFrom(in, classInfo);
            loaded = true;
        } 
        catch (Exception e) { throw new Error(e.toString()); }
    }
    public String toString() { 
        if (!loaded) createFromData();

        String str = "";
        for (int i=0; i<tableLength; i++) 
            str += "\n    " + localVariableTable[i];
        return str;
    }
}  


/////////////////////////
class ExceptionTableEntry {
    JavaClass  classInfo;
    public int      start_pc;  
    public int      end_pc;  
    public int      handler_pc;  
    public int      catch_type;  

    public int getStartPC  () { return start_pc;   }
    public int getEndPC    () { return end_pc;     }
    public int getHandlerPC() { return handler_pc; }
    public String getCatchType() { return classInfo.constantAtIndex(catch_type).toString(); }

    static ExceptionTableEntry createFrom(DataInputStream in, JavaClass cls) {
        ExceptionTableEntry c = new ExceptionTableEntry();
        try {
            c.classInfo  = cls;
            c.start_pc   = in.readUnsignedShort();
            c.end_pc     = in.readUnsignedShort();
            c.handler_pc = in.readUnsignedShort();
            c.catch_type = in.readUnsignedShort();
        } catch (Exception e) { throw new Error(e.toString()); }
        return c;
    }
    public String toString() { 
        String str = "exception " + getCatchType()
                   + " starts at code[" + handler_pc
                   + "]; handles code[" + start_pc
                   + "] - code[" + end_pc
                   + "]";
        return str;
    }
}
class LineNumberTableEntry {
    JavaClass  classInfo;
    int      start_pc;  
    int      line_number;  

    public int getStartPC   () { return start_pc; }
    public int getLineNumber() { return line_number; }

    static LineNumberTableEntry createFrom(DataInputStream in, JavaClass cls) {
        LineNumberTableEntry c = new LineNumberTableEntry();
        try {
            c.classInfo   = cls;
            c.start_pc    = in.readUnsignedShort();
            c.line_number = in.readUnsignedShort();
        } catch (Exception e) { throw new Error(e.toString()); }
        return c;
    }
    public String toString() { return "code[" + start_pc + "] is line " + line_number; }
}
class LocalVariableTableEntry {
    JavaClass  classInfo;
    int      start_pc;  
    int      length;  
    int      nameIndex;  
    int      sigIndex;  
    int      slot;  

    public int getStartPC() { return start_pc; }
    public int getLength () { return length; }
    public int getSlot   () { return slot; }
    public String getVariableName     () { return classInfo.constantAtIndex(nameIndex).toString(); }
    public String getVariableSignature() { return classInfo.constantAtIndex(sigIndex ).toString(); }


    static LocalVariableTableEntry createFrom(DataInputStream in, JavaClass cls) {
        LocalVariableTableEntry c = new LocalVariableTableEntry();
        try {
            c.classInfo = cls;
            c.start_pc  = in.readUnsignedShort();
            c.length    = in.readUnsignedShort();
            c.nameIndex = in.readUnsignedShort();
            c.sigIndex = in.readUnsignedShort();
            c.slot           = in.readUnsignedShort();
        } catch (Exception e) { throw new Error(e.toString()); }
        return c;
    }
    public String toString() { 
        String str = "var# " + slot
                   + ": " + getVariableSignature()
                   +  " " + getVariableName()
                   + "  scope: code[" + start_pc
                   + "] - code[" + (start_pc + length)
                   + "]";

        return str;
    }
}
