//////////////////////license & copyright header///////////////////////
//                                                                   //
//              Java Cannery - jar management utility.               //
//                Copyright (c) 1998 by Kevin Kelley                 //
//                                                                   //
// This program is free software; you can redistribute it and/or     //
// modify it under the terms of the GNU General Public License as    //
// published by the Free Software Foundation; either version 2 of    //
// the License, or (at your option) any later version.               //
//                                                                   //
// This program is distributed in the hope that it will be useful,   //
// but WITHOUT ANY WARRANTY; without even the implied warranty of    //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     //
// GNU General Public License for more details.                      //
//                                                                   //
// You should have received a copy of the GNU General Public License //
// along with this program in the file 'gpl.html'; if not, write to  //
// the Free Software Foundation, Inc., 59 Temple Place - Suite 330,  //
// Boston, MA 02111-1307, USA, or contact the author:                //
//                                                                   //
//                       Kevin Kelley  <kelley@iguana.ruralnet.net>  //
//                                                                   //
////////////////////end license & copyright header/////////////////////
package starlight.java;
import java.io.*;
import java.util.*;
import java.util.zip.*;
import starlight.util.*;

/** A JarArchive is a tree-oriented group of resources, accessed
  * by qualified-name (relative path + '/' + filename).  It maintains
  * a manifest file; resources that are added to the archive get
  * entries in the manifest.  When loading an existing archive, the
  * manifest it contains, if any, is used, unmodified.  Further
  * additions to the archive will update it.  This allows the user
  * to edit/strip/add to the manifest without having the changes
  * overwritten, while also keeping it updated.
  */
public class JarArchive extends ResourceTree {

  public static String MANIFEST = Manifest.MANIFEST; //"META-INF/MANIFEST.MF"
  public String zipComment = "created by Starlight";
  private int compMethod = ZipOutputStream.DEFLATED;
  private int compLevel = Deflater.DEFAULT_COMPRESSION;

    private boolean useSHA = true;
    private boolean useMD5 = true;

    String    name;         // full pathname of jar
    ZipFile   archive;      // leave entry data cached in zip until needed
    ResourceTree parent;    // if we're an entry in another tree...
    Manifest  manifest;     // loaded from jar, and added to on addResource()
    boolean   modified;
    boolean   createDirs = true;    // should create entries for dirs?

    /** Construct from a file.  Leave entry data cached in the file
      * until needed.
      */
    public JarArchive(File f)
        {   try {
                name = f.getAbsolutePath();
                File src = f;
                if (src.exists()) {
                    archive = new ZipFile(src);
                    scan(archive);
                } else
                    dirTree = new DirectoryTree(name, separatorChar);

            } catch (IOException e) {
                Msg.error(e.toString()+" creating JarArchive from "+f.getPath());
            }
        }
    /** Construct from an input stream.  This loads all entry data
      * into memory (kept in the dirtree member).
      */
    public JarArchive(ResourceTree parent, String name, InputStream in)
        { this.parent = parent; this.name = name; buildFrom(in); }


    public String getName() { return name; }

    public void setName(String name) { this.name = name; modified = true; }

    public boolean maintainManifest() { return true; }

    /** specify whether to include SHA digest for manifest entries */
    public void setIncludeSHADigest(boolean b)
        { if (manifest!=null) manifest.setIncludeSHADigest(useSHA=b); }

    /** specify whether to include MD5 digest for manifest entries */
    public void setIncludeMD5Digest(boolean b)
        { if (manifest!=null) manifest.setIncludeMD5Digest(useMD5=b); }

    public String[] getDirContents(String subdir) {
        return dirTree.list(subdir);
    }
    // !!!implement filtered list!!!
    public String[] getDirContents(String subdir, FilenameFilter filt) {
        return dirTree.list(subdir);
    }

    public Enumeration childNames() {
        return new Enumeration() {
            Stack unscanned = new Stack();

            // seed the stack with the 'root' contents
            int xxx = pushSubdirs("");

            private int pushSubdirs(String dir) {
                String[] contents = dirTree.list(dir);
                for (int i=contents.length-1; i>=0; --i)
                    unscanned.push(contents[i]);
                return contents.length;
            }
            public boolean hasMoreElements() { return !unscanned.empty(); }

            public Object nextElement() {
                String next = (String) unscanned.pop();
                if (isDirectory(next)) {
                    String[] contents = dirTree.list(next);
                    for (int i=contents.length-1; i>=0; --i)
                        unscanned.push(next + contents[i]);
                } return next;
        }   };
    }

    public boolean isDirectory(String qualName) {
        return qualName != null && qualName.charAt(qualName.length()-1) == '/';
    }
    public boolean contains(String qualName)
        { return (dirTree.get(qualName) != null); }

    public long lastModified(String qualName) {
        Resource res = getResource(qualName);
        if (res == null) return -1;
        return res.lastModified();
    }
    public byte[] getBytes(String qualName) {
        Resource res = getResource(qualName);
        if (res == null) return null;
        return res.getContent();
    }

    public Resource getResource(String qualName) {
        DirectoryTree.Node node = dirTree.get(qualName);
        if (node != null) {
            Object val = node.value;
            if (val instanceof Resource) {
                return (Resource) val;
            }
            else if (val instanceof ZipEntry && archive != null) {
                ZipEntry entry = (ZipEntry) val;
                byte[] data = null;
                try { data = readFully(archive.getInputStream(entry),
                                       entry.getSize());
                } catch (IOException e) {Msg.error("reading zip entry "+qualName);}
                node.value = new Resource(entry, data);
                return (Resource) node.value;
            }
        }
        if (!isDirectory(qualName))
            Msg.error("couldn't get resource " + qualName);
        return null;
    }

    public void addResource(Resource r) {
        if (contains(r.getName()))
            removeResource(r.getName());

        dirTree.add(r.getName(), r);

        if (maintainManifest()) {
            if (manifest == null)
                manifest = new Manifest();
            manifest.setIncludeSHADigest(useSHA);
            manifest.setIncludeMD5Digest(useMD5);
            manifest.addSectionForResource(r);
        }
        modified = true;
        //notifyObservers();
    }

    /** removing a node that is a directory will remove all its
      * contents, in one swell foop.
      */
    public void removeResource(String name) {
        DirectoryTree.Node node = dirTree.get(name);
        if (node != null) {
            if (maintainManifest() && manifest!=null) {
                if (isDirectory(name)) {
                    Stack stack = new Stack();
                    stack.push(name);
                    while (!stack.empty()) {
                        String parent = (String) stack.pop();
                        String[] list = getDirContents(parent);
                        for (int i=0; list!=null && i<list.length; i++) {
                            String child = parent + list[i];
                            if (isDirectory(child))
                                stack.push(child);
                            else manifest.removeManifestSection(child);
                    }   }
                } else manifest.removeManifestSection(name);
            }

            dirTree.remove(name);

            modified = true;
            //notifyObservers();

        } else if ( isDirectory(name) ) {
            /*non-existent node is a directory; ignore the remove request*/
        } else
            throw new Error(name+" not present; can't remove!");
    }

    // create an output stream on which to save a child resource
    // (return a do-nothing stream.  Jar entries can't be saved
    // individually; we have to mark modified, and write the whole
    // jar out later).
    protected OutputStream getOutputStream(String qualName) {
        modified = true;
        return new OutputStream() {
            public void write(int b) {}
            public void write(byte[] b, int off, int len) {}
        };
    }

    ////////////////////////////////////////////
    // get an output stream suitable for writing this jar onto.
    protected OutputStream getOutputStream() throws IOException {
        if (parent != null)
            return parent.getOutputStream(getName());
        return new FileOutputStream(new File(name));
    }

    public void setCompression(boolean compress)
        { if (compress) compMethod = ZipOutputStream.DEFLATED;
          else          compMethod = ZipOutputStream.STORED;
          modified = true;
          notifyObservers();
        }

    public boolean isModified() { return modified; }

    public void write() {

        load();     // force read of ZipEntries into memory

        OutputStream out = null;
        ZipOutputStream zip = null;
        String tempName = name + ".tempcopy";
        try {
            // if we've got a parent, write directly to whatever
            // stream it gives us.  Otherwise, write to a temp file
            // and on success, rename it.
            boolean directWrite = (parent != null);
            File tempFile = null;
            if (directWrite)
                out = parent.getOutputStream(name);
            else {
                tempFile = new File(tempName);
                out = new FileOutputStream(tempFile);
            }
            zip = new ZipOutputStream(out);
            zip.setComment(zipComment);
            zip.setMethod(compMethod);

            if (manifest != null)  {
                manifest.flush();
                zip.putNextEntry(manifest.asZipEntry());
                zip.write(manifest.getData());
                zip.closeEntry();
            }

            for (Enumeration e=childNames(); e.hasMoreElements(); ) {
                String elt = (String) e.nextElement();
                Msg.notify("fetching "+elt);
                Resource res = getResource(elt);
                if (res==null && isDirectory(elt) && createDirs) {
                    res = new Resource(elt, new byte[0]);
                }
                if (res==null && !(isDirectory(elt) && !createDirs))
                    throw new Error("missing resource "+elt);
                if (res != null) {  // if !createDirs, skip missing dir resources
                    ZipEntry entry = res.asZipEntry();
                    byte[] data = res.getContent();
                    zip.putNextEntry(entry);
                    if (data != null) zip.write(data);
                    zip.closeEntry();
                }
            }
            zip.close();

            if (archive!=null) {    // now that all entries have been read,
                archive.close();    // no need to keep zip archive around.
                archive = null;     // all its contents are in memory.
            }
            if (!directWrite) {
                // creation of new (temp) archive has succeeded; now
                // kill the old and rename the temp.
                File file = new File(name);
                try {
                    if (file.exists())
                        file.delete();
                    tempFile.renameTo(file);
                } catch (Exception e) {
                    Msg.error("copy failed: unable to replace old '"
                        + file.getPath()
                        + "' with the temporary '"+tempName+"'.");
                }
            }
            modified = false;
        }
        catch (Throwable e) {
            Msg.error(e.toString() + ": write failed!");
        } finally {
                 if (zip!=null) try { zip.close(); } catch (Throwable ex) {}
            else if (out!=null) try { out.close(); } catch (Throwable ex) {}
        }
        notifyObservers();
    }

    // build the directory tree and stash the zipEntries for later
    public void scan(ZipFile zip) {
        manifest = null;
        dirTree = new DirectoryTree(getName(), separatorChar);

        for (Enumeration e=zip.entries(); e.hasMoreElements(); ) {
            ZipEntry entry = (ZipEntry) e.nextElement();

            if (MANIFEST.equals(entry.getName())) {
                try {
                    InputStream in = zip.getInputStream(entry);
                    byte[] data = readFully(in, entry.getSize());
                    manifest = new Manifest(data, entry.getTime());
                } catch (IOException x) { Msg.error(x+" reading manifest!"); }
            }
            else dirTree.add(entry.getName(), entry);
        }
    }

    // resolve any ZipEntries by reading them from the archive, then
    // close the archive.  Forces full jar into memory.
    public void load() {
        try {
            for (Enumeration e=childNames(); e.hasMoreElements(); ) {
                String elt = (String) e.nextElement();
                Msg.notify("fetching "+elt);
                Resource res = getResource(elt);
            }
            archive.close();  // presume everybody's loaded
            archive = null;
        } catch (Throwable e) {
            Msg.error(e.toString() + " loading archive '"+getName()+"'!");
        }
    }


    // build the directory tree, and load the resources
    public void buildFrom(InputStream in) {
        dirTree = new DirectoryTree(getName(), separatorChar);
        ZipInputStream zip = null;
        try {
            zip = new ZipInputStream(in);

            ZipEntry zipEntry = zip.getNextEntry();

            if (zipEntry != null && MANIFEST.equals(zipEntry.getName())) {
                manifest = new Manifest(readFully(zip, zipEntry.getSize()),
                                        zipEntry.getTime());
                zip.closeEntry();
                zipEntry = zip.getNextEntry();
            }
            else manifest = null;

            while (zipEntry != null) {
                Resource res = new Resource(zipEntry,
                                            readFully(zip, zipEntry.getSize()));
                dirTree.add(res.getName(), res);
                zip.closeEntry();
                zipEntry = zip.getNextEntry();
            }
            zip.close();
        } catch (Throwable e) {
                 if (zip!=null) try { zip.close(); } catch (Throwable ex) {}
            else if ( in!=null) try {  in.close(); } catch (Throwable ex) {}
            Msg.error(e.toString());
        }
        notifyObservers();
    }
    /** specialized version of ByteArray.readFully.  InputStreams created
      * from zip files don't know their correct available() size, so this
      * lets it be controlled by the value in the ZipEntry, if it is set.
      *  This might speed things up somewhat.
      */
    byte[] readFully(InputStream in, long estimatedSize) throws IOException {
        if (estimatedSize == -1) estimatedSize = 16000;
        ByteArray ba = new ByteArray((int)estimatedSize);
        ba.appendFrom(in, (int)estimatedSize);
        ba.trimToSize();
        return ba.getArrayRep();
    }
}
