//////////////////////license & copyright header///////////////////////
//                                                                   //
//              Java Cannery - jar management utility.               //
//                Copyright (c) 1998 by Kevin Kelley                 //
//                                                                   //
// This program is free software; you can redistribute it and/or     //
// modify it under the terms of the GNU General Public License as    //
// published by the Free Software Foundation; either version 2 of    //
// the License, or (at your option) any later version.               //
//                                                                   //
// This program is distributed in the hope that it will be useful,   //
// but WITHOUT ANY WARRANTY; without even the implied warranty of    //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     //
// GNU General Public License for more details.                      //
//                                                                   //
// You should have received a copy of the GNU General Public License //
// along with this program in the file 'gpl.html'; if not, write to  //
// the Free Software Foundation, Inc., 59 Temple Place - Suite 330,  //
// Boston, MA 02111-1307, USA, or contact the author:                //
//                                                                   //
//                       Kevin Kelley  <kelley@iguana.ruralnet.net>  //
//                                                                   //
////////////////////end license & copyright header/////////////////////
package starlight.java;
import java.io.*;
import java.util.*;
import starlight.util.*;

/**
*   With respect to this class, JavaClassClosure, any class name
* referenced is expected to be a fully-qualified (pkg.subpkg.className)
* class name, and is not to include the file-system's '.class'
* extension.
*/
public class JavaClassClosure {

    //   key: JavaClass
    // value: Hashtable(JavaClassName, JavaClass)
    private Hashtable classClosure = new Hashtable();

    private Hashtable excludedRoots = new Hashtable();


    public JavaClassClosure() {}


    /**
    *   Add a root directory to the list of packages to ignore
    */
    public void excludeRoot(String root) {
        root = root.replace(File.separatorChar, '/');
        root = root.replace('/', '.');
        // forestall confusion if a class name matches a pkg
        if (! root.endsWith("."))
            root += ".";
        excludedRoots.put(root, new File(root));
    }

    /**
    *   Lookup and add the specified class.
    */
    public void add(String className)
        { add(JavaClass.forName(className)); }

    /**
    *   Add a class by closing over its dependencies, and entering
    *  it and its closure into the classClosure table.
    */
    public void add(JavaClass cls)
        { closeOver(cls); }

    /**
    *   Returns Enumeration of JavaClass - all that are reachable
    *   from one or more added initial classes.
    */
    public Enumeration elements() { return classClosure.keys(); }

    /**
    *   Returns Enumeration of JavaClass's directly referenced
    *   by the given class.
    */
    public Enumeration getClassesReferencedBy(String className) {
        JavaClass cls = JavaClass.forName(className);
        Hashtable referents = (Hashtable) classClosure.get(cls);
        if (referents == null)
            return new EmptyEnumeration();
        return referents.elements();
    }


    //   Return true if a qualified classname is a child of
    //  one of the excluded roots.
    private boolean isExcluded(String qualClass) {
        Enumeration e = excludedRoots.keys();
        while (e.hasMoreElements()) {
            String root = (String) e.nextElement();
            if (qualClass.startsWith(root))
                return true;
        }
        return false;
    }

    private void closeOver(JavaClass cls) {
        Stack stack = new Stack();
        stack.push(cls);
        while (!stack.empty()) {
            JavaClass current = (JavaClass) stack.pop();
            if ( ! classClosure.containsKey(current) ) {
                Hashtable currentRefs = getReferencedClasses(current);
                classClosure.put(current, currentRefs);
                //System.out.println("adding "
                //    +current.getQualifiedClassName()+" to closure");

                for (Enumeration e=currentRefs.elements(); e.hasMoreElements(); )
                    stack.push((JavaClass)e.nextElement());
        }   }
    }

    /**
    *   Return a hashtable (key: String className; value: JavaClass cls;)
    *   containing all the classes directly referred to by the passed class.
    *
    *   o   thisClass    - CONSTANT_Class
    *   o   superClass   - CONSTANT_Class
    *   o   interfaces[] - CONSTANT_Class
    *
    *   references from code:
    *   o   CONSTANT_Fieldref      -
    *   o   CONSTANT_Methodref     -
    *   o   CONSTANT_IntfMethodref - all contain reference to CONSTANT_Class
    *
    *   signature descriptors:
    *   o   CONSTANT_NameAndType  - referenced class will contain fields/methods
    *                               matching these signature descriptors.
    *
    *   fields[] / methods[]:
    *   o   name - identifier
    *   o   type - signature descriptor string
    *
    *   Attributes:
    *   o   method[].attributes[].ExceptionsAttribute - in method code; contains
    *                                                   array of CONSTANT_Class for
    *                                                   "throws" clause.
    *   o   method[].attributes[].LineNumberTable     - n/a
    *   o   method[].attributes[].LocalVariableTable  - name/descriptor (utf8) for
    *                                                   local variables.
    *                                                   Redundant; optional.  I hope.
    */
    private Hashtable getReferencedClasses(JavaClass cls) {
        Hashtable table = new Hashtable();
        //System.out.println("closing over " + cls.getQualifiedClassName());

        for (int i=0; i<cls.numConstants(); i++) {
            ConstantInfo con = cls.constantAtIndex(i);
            if ((con != null) && (con.tag == ConstantInfo.CONST_Class)) {
                String clsName = ((CONSTANT_Class)con).getClassName();
                if (clsName.indexOf('[') >= 0) {
                    Signature sig = Signature.createFrom(clsName);
                    clsName = sig.basicTypeName();
                }
                if (!isBuiltIn(clsName)
                 && !isExcluded(clsName)
                 && !table.containsKey(clsName)) {
                    //System.out.println("adding from class ref " + clsName);
                    JavaClass refClass = JavaClass.forName(clsName);
                    table.put(clsName, refClass);
                } //else System.out.println("skipping class ref " + clsName);
        }   }

        for (int i=0; i<cls.numFields; i++) {
            Signature sig = cls.fields[i].getSignature();
            if (!sig.isBuiltIn()) {
                String name = sig.basicTypeName();
                if (!isExcluded(name) && !table.containsKey(name)) {
                    //System.out.println("adding from field ref " + name);
                    JavaClass fldType = JavaClass.forName(name);
                    table.put(name, fldType);
        }   }   }

        for (int i=0; i<cls.numMethods; i++) {
            JavaMethod meth = cls.methods[i];
            MethodSignature sig = meth.getMethodSignature();
            if (!sig.getReturnType().isBuiltIn()) {
                String name = sig.getReturnType().basicTypeName();
                if (!isExcluded(name) && !table.containsKey(name)) {
                    //System.out.println("adding from return ref " + name);
                    JavaClass rtnType = JavaClass.forName(name);
                    table.put(name, rtnType);
            }   }
            Signature[] parms = sig.getParameters();
            for (int j=0; j<parms.length; j++) {
                if (!parms[j].isBuiltIn()) {
                    String name = parms[j].basicTypeName();
                    if (!isExcluded(name) && !table.containsKey(name)) {
                        //System.out.println("adding from param ref " + name);
                        JavaClass parm = JavaClass.forName(name);
                        table.put(name, parm);
        }   }   }   }
        return table;
    }
static Hashtable builtins;
static {
    builtins = new Hashtable();
    builtins.put(   "void",      Void.TYPE);
    builtins.put("boolean",   Boolean.TYPE);
    builtins.put(   "byte",      Byte.TYPE);
    builtins.put(   "char", Character.TYPE);
    builtins.put(  "short",     Short.TYPE);
    builtins.put(    "int",   Integer.TYPE);
    builtins.put(   "long",      Long.TYPE);
    builtins.put(  "float",     Float.TYPE);
    builtins.put( "double",    Double.TYPE);
}
static boolean isBuiltIn(String name) {
    return builtins.containsKey(name);
}
}
