//////////////////////license & copyright header///////////////////////
//                                                                   //
//              Java Cannery - jar management utility.               //
//                Copyright (c) 1998 by Kevin Kelley                 //
//                                                                   //
// This program is free software; you can redistribute it and/or     //
// modify it under the terms of the GNU General Public License as    //
// published by the Free Software Foundation; either version 2 of    //
// the License, or (at your option) any later version.               //
//                                                                   //
// This program is distributed in the hope that it will be useful,   //
// but WITHOUT ANY WARRANTY; without even the implied warranty of    //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     //
// GNU General Public License for more details.                      //
//                                                                   //
// You should have received a copy of the GNU General Public License //
// along with this program in the file 'gpl.html'; if not, write to  //
// the Free Software Foundation, Inc., 59 Temple Place - Suite 330,  //
// Boston, MA 02111-1307, USA, or contact the author:                //
//                                                                   //
//                       Kevin Kelley  <kelley@iguana.ruralnet.net>  //
//                                                                   //
////////////////////end license & copyright header/////////////////////
package starlight.java;

import java.io.*;
import java.util.*;
import starlight.io.*;
import starlight.util.Base64;
import starlight.util.Msg;

/** Manifest is a resource in a jar file that contains a catalog
  * of other resources in the jar, in the form of 'sections' of
  * key/value pairs, each section representing one resource in
  * the jar.  Resources don't necessarily all have sections.
  *  Manifest object can be created from preexisting manifest file,
  * in which case the file will be parsed into sections.
  *  Additional sections can be added; method
  * <CODE>addEntryForResource(Resource)</code>
  * appends a new section to the
  * manifest, with following key/value pairs:
  *
  *    Name: pkg/filename.class
  *    Digest-Algorithms: SHA, MD5
  *    SHA-Digest: base64-encoded digest
  *    MD5-Digest: base64-encoded digest
  *
  */
public class Manifest extends Resource {

  public static String MANIFEST = "META-INF/MANIFEST.MF";

    Vector sections = new Vector();

    boolean useSHA = true;
    boolean useMD5 = true;

    public Manifest() { super(MANIFEST); }

    public Manifest(byte[] data, long modTime) {
        super(MANIFEST, data, modTime);
        parseFromBytes(data);
    }

    /** specify whether to include SHA digest for manifest entries */
    public void setIncludeSHADigest(boolean b) { useSHA = b; }

    /** specify whether to include MD5 digest for manifest entries */
    public void setIncludeMD5Digest(boolean b) { useMD5 = b; }

    public class Pair {
        String key;
        String value;
        Pair(String k, String v) {key=k; value=v; }
    }

    public class ManifestEntry {

        Vector headers = new Vector();  //Vector<Pair>

        public ManifestEntry() {}

        public ManifestEntry(String name) {
            setName(name);
        }

        /** add a header key/value pair.  Doesn't allow duplicate (nonempty)
          * keys.
          */
        public void addHeader(String k, String v) {
            for (Enumeration e=headers.elements(); e.hasMoreElements(); ) {
                Pair pair = (Pair) e.nextElement();
                if (pair.key.length()>0 && pair.key.equals(k)) {
                    pair.value = v;
                    return;
            }    }
            headers.addElement(new Pair(k, v));
        }
        /** returns value for key.  Used on empty keys, will return first
          * unkeyed value.
          */
        public String getHeaderValue(String key) {
            for (Enumeration e=headers.elements(); e.hasMoreElements(); ) {
                Pair pair = (Pair) e.nextElement();
                if (pair.key.equals(key))
                    return pair.value;
            } return null;
        }

        public String getName() { return getHeaderValue("Name"); }
        public void setName(String name) { addHeader("Name", name); }

        /** Returns the SHA-Digest header's value, decoded to byte[]
          * if it is found; null otherwise.
          */
        public byte[] getSHA() {
            String s = getHeaderValue("SHA-Digest");
            if (s==null) return null;
            return Base64.decode(s.toCharArray());
        }
        /** Adds a key/value header for the provided SHA digest,
          * properly encoded to Base64.
          */
        public void setSHA(byte[] sha) {
            String s = getHeaderValue("Digest-Algorithms");
            if (s==null || s.length()==0)
                 s  = "SHA";
            else s += ", SHA";
            addHeader("Digest-Algorithms", s);
            addHeader("SHA-Digest", new String(Base64.encode(sha)));
        }

        /** Returns the MD5-Digest header's value, decoded to byte[]
          * if it is found; null otherwise.
          */
        public byte[] getMD5() {
            String s = getHeaderValue("MD5-Digest");
            if (s==null) return null;
            return Base64.decode(s.toCharArray());
        }
        /** Adds a key/value header for the provided MD5 digest,
          * properly encoded to Base64.
          */
        public void setMD5(byte[] md5) {
            String s = getHeaderValue("Digest-Algorithms");
            if (s==null || s.length()==0)
                 s  = "MD5";
            else s += ", MD5";
            addHeader("Digest-Algorithms", s);
            addHeader("MD5-Digest", new String(Base64.encode(md5)));
        }

        /** break line into key: value, if possible; otherwise store
          * it with empty key.
          */
        public void addHeader(String line) {
            int ix = line.indexOf(':');
            if (ix>0 && ix<line.length()-1 && line.charAt(ix+1)==' ')
                addHeader(line.substring(0, ix), line.substring(ix+2));
            else addHeader("", line);
        }

        void printOn(PrintWriter print) {
            for (Enumeration e=headers.elements(); e.hasMoreElements(); ) {
                Pair pair = (Pair) e.nextElement();
                if (pair.key.length()>0)
                     print.println(pair.key + ": " + pair.value);
                else print.println(pair.value);
            }
        }
    }



    /** Create a manifest entry for the resource.
      * Overwrites any existing entry.
      */
    public void addSectionForResource(Resource res) {
        ManifestEntry entry = new ManifestEntry(res.getName());
        if (useSHA) entry.setSHA(res.getSHA());
        if (useMD5) entry.setMD5(res.getMD5());
        sections.addElement(entry);
    }
    public void removeManifestSection(String res) {
        for (int i=0; i<sections.size(); i++) {
            ManifestEntry entry = (ManifestEntry) sections.elementAt(i);
            if (res.equals(entry.getName())) {
                sections.removeElementAt(i);
                return;
    }   }   }

    public void flush() {
        createManifestBytes();
    }


int numSections() { return sections.size(); }
ManifestEntry getSection(int i) {
    if (numSections() < i)
        return null;
    return (ManifestEntry) sections.elementAt(i);
}

/** Update the manifest data to represent current contents
  */
private void createManifestBytes() {
    ByteArrayOutputStream byteStream;
    PrintWriter print = new PrintWriter(
                        new ContinuatorOutputStream( byteStream =
                        new ByteArrayOutputStream(16384)));

    ManifestEntry first = getSection(0);
    if (first==null || first.getHeaderValue("Manifest-version")==null) {
        first = new ManifestEntry();
        first.addHeader("Manifest-version", "1.0");
        sections.insertElementAt(first, 0);
    }

    for (Enumeration e=sections.elements(); e.hasMoreElements(); ) {
        ManifestEntry entry = (ManifestEntry) e.nextElement();

        entry.printOn(print);
        if (e.hasMoreElements())
            print.println("");
    }
    print.close();
    setData(byteStream.toByteArray());
    setModifiedTime((new Date()).getTime());
}

/** manifest file consists of sections, separated by blank lines.
  * First 'section' is (should be) 'Manifest-version: 1.0', with
  * an optional second 'Minimum-version: ' line.
  * Remaining, normal, sections will generally start with a
  * 'name: whatever/file.class' header.  In any case, each blank-
  * line-delimited section will get its own ManifestEntry.  If
  * there is no name: header, it will be kept as an unnamed section
  * so as to maintain any extra info in the file (comments, etc).
  *  Multiple blank lines, however, will be conflated down to one.
  */
private void parseFromBytes(byte[] data) {
    BufferedReader in = new BufferedReader(
                        new InputStreamReader(
                        new ContinuatorInputStream(
                        new ByteArrayInputStream(data) )));
    String line = null;
    ManifestEntry section = null;

    try {
        while ((line=in.readLine()) != null) {
            if (line.trim().length() == 0) {
                // blank line; end section and ignore.
                if (section != null) {
                    sections.addElement(section);
                    section = null;
            }   } else {
                // non-blank line; add it to a section.
                if (section == null)
                    section = new ManifestEntry();
                section.addHeader(line);
        }   }
        if (section != null) {
            sections.addElement(section);
            section = null;
        }
    }
    catch (IOException e) { Msg.error(e+" parsing manifest!"); }
    finally { try { if (in!=null) in.close(); } catch(Exception e) {} }
}
}
