//////////////////////license & copyright header///////////////////////
//                                                                   //
//              Java Cannery - jar management utility.               //
//                Copyright (c) 1998 by Kevin Kelley                 //
//                                                                   //
// This program is free software; you can redistribute it and/or     //
// modify it under the terms of the GNU General Public License as    //
// published by the Free Software Foundation; either version 2 of    //
// the License, or (at your option) any later version.               //
//                                                                   //
// This program is distributed in the hope that it will be useful,   //
// but WITHOUT ANY WARRANTY; without even the implied warranty of    //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     //
// GNU General Public License for more details.                      //
//                                                                   //
// You should have received a copy of the GNU General Public License //
// along with this program in the file 'gpl.html'; if not, write to  //
// the Free Software Foundation, Inc., 59 Temple Place - Suite 330,  //
// Boston, MA 02111-1307, USA, or contact the author:                //
//                                                                   //
//                       Kevin Kelley  <kelley@iguana.ruralnet.net>  //
//                                                                   //
////////////////////end license & copyright header/////////////////////
package starlight.java;
import java.io.*;
import java.util.*;
import java.util.zip.*;
import java.text.*;
import java.security.*;
import starlight.util.*;

public class Resource {

  String name;
  byte[] entryData;
  ZipFile dataStash;
  String contentType = "unknown";
  String comment = "";
  long modTime = -1;
  long csize = -1;

  byte[] shaDigest;   // cache the digests; only calc once.
  byte[] md5Digest;
  long crc32;

    public Resource(File file) throws IOException {
        this(file.getPath(),
             ByteArray.readFully(new FileInputStream(file)),
             file.lastModified());
    }
    public Resource(JavaClass cls) {
        this(cls.getClassFileName(),
             cls.getClassBytes(),
             cls.lastModified());
    }
    // prolly not a good idea
    /*public*/ Resource(ZipEntry entry, ZipFile zip) {
        this(entry, (byte[])null);
        dataStash = zip;
    }
    public Resource(ZipEntry entry, byte[] data) {
        this(entry.getName(), data, entry.getTime());
        csize = entry.getCompressedSize();
        comment = entry.getComment();
    }
    public Resource(String name, byte[] data, long mod)
        { this(name, data); setModifiedTime(mod); }

    public Resource(String name, byte[] data)
        { this(name); setData(data); }

    public Resource(String name)
        { this.name = name.replace(File.separatorChar, '/'); }

    public ZipEntry asZipEntry() {
        ZipEntry entry = new ZipEntry(name);
        entry.    setCrc(getCrc()      );
        entry.   setSize(getSize()     );
        entry.   setTime(lastModified());
        entry.setComment(getComment()  );
        return entry;
    }

    public void setName(String name)
    {   if (dataStash!=null)
            getData();
        this.name=name.replace(File.separatorChar, '/');
    }

    public String         getName() { return name;      }
    public String    getShortName() { return name.substring(   name.lastIndexOf('/')+1); }
    public String         getPath() { return name.substring(0, name.lastIndexOf('/')+1); }

    public long           getSize() { return (entryData!=null)? entryData.length: -1; }
    public long getCompressedSize() { return csize;     }

    public long      lastModified() { return modTime;   }

    public String     getDateTime() { return DATE_TIME.format(new Date(modTime)); }
    public String         getDate() { return      DATE.format(new Date(modTime)); }
    public String         getTime() { return      TIME.format(new Date(modTime)); }

    public byte[]      getContent() { return getData(); }
    public byte[] getData() {
        if (entryData == null && dataStash != null) {
            try {
            ZipEntry entry = dataStash.getEntry(getName());
            InputStream in = dataStash.getInputStream(entry);
            setData(ByteArray.readFully(in));
            in.close();
            dataStash = null;
            } catch (IOException e) { Msg.error("getting data for "+name); }
        }
        return entryData;
    }

    public String      getComment() { return comment;   }

    public void setModifiedTime(long t) { modTime = t;  }
    public void    setComment(String c) { comment = c;  }

    public void setData(byte[] d) {
        entryData = d;
        dataStash = null;
        shaDigest = null;
        md5Digest = null;
        crc32     = 0;
        csize     = -1;
    }

    public long getCrc() {
        if (crc32 == 0) { // if crc actually is 0, we'll recalc it every time
            crcCalculator.reset();
            crcCalculator.update(getData());
            crc32 = crcCalculator.getValue();
        } return crc32;
    }
    public byte[] getSHA() {
        if (shaDigest == null && shaCalculator != null)
            shaDigest = shaCalculator.digest(getData());
        return shaDigest;
    }
    public byte[] getMD5() {
        if (md5Digest == null && md5Calculator != null)
            md5Digest = md5Calculator.digest(getData());
        return md5Digest;
    }


  static String CONTENT_TYPES = "content-types.properties";
  static Properties contentTypes;
  static {
    Properties defaults = new Properties();
    try {
        FileInputStream in = new FileInputStream(CONTENT_TYPES);
        defaults.load(in);
        in.close();
    } catch (FileNotFoundException e) {
        //defaults.put("Verbose", "true");
    } catch (IOException e)
    { System.err.println(e.toString()
        + " loading properties file " + CONTENT_TYPES);
    }
    contentTypes = new Properties(defaults);
  }

  static private CRC32 crcCalculator = new CRC32();
  static private MessageDigest shaCalculator;
  static private MessageDigest md5Calculator;
  static {
    try { shaCalculator = MessageDigest.getInstance("SHA-1");
    } catch (NoSuchAlgorithmException e) {}
    try { md5Calculator = MessageDigest.getInstance("MD5");
    } catch (NoSuchAlgorithmException e) {}
  }

  static DateFormat DATE_TIME;
  static DateFormat DATE;
  static DateFormat TIME;
  static {
    DATE_TIME = DateFormat.getDateTimeInstance(DateFormat.SHORT, DateFormat.SHORT);
    DATE = DateFormat.getDateInstance(DateFormat.SHORT);
    TIME = DateFormat.getTimeInstance(DateFormat.SHORT);
    DATE_TIME.setTimeZone(TimeZone.getDefault());
    DATE.setTimeZone(TimeZone.getDefault());
    TIME.setTimeZone(TimeZone.getDefault());
    if (DATE_TIME instanceof SimpleDateFormat)
        ((SimpleDateFormat)DATE_TIME).applyPattern("h:mm a  dd MMM ''yy");
    if (DATE instanceof SimpleDateFormat)
        ((SimpleDateFormat)DATE).applyPattern("d MMM ''yy");
    if (TIME instanceof SimpleDateFormat)
        ((SimpleDateFormat)TIME).applyPattern("h:mm a");
  }

  public static String shortDateTimeFormat(long time)
    { return DATE_TIME.format(new Date(time)); }

  public static String shortDateFormat(long time)
    { return DATE.format(new Date(time)); }

  public static String shortTimeFormat(long time)
    { return TIME.format(new Date(time)); }

  public String toString() {
    return "Resource: " + name + " ("
        + ((entryData!=null)
            ? (String.valueOf(entryData.length) + " bytes)" )
            : "no data)");
  }
}
