//////////////////////license & copyright header///////////////////////
//                                                                   //
//              Java Cannery - jar management utility.               //
//                Copyright (c) 1998 by Kevin Kelley                 //
//                                                                   //
// This program is free software; you can redistribute it and/or     //
// modify it under the terms of the GNU General Public License as    //
// published by the Free Software Foundation; either version 2 of    //
// the License, or (at your option) any later version.               //
//                                                                   //
// This program is distributed in the hope that it will be useful,   //
// but WITHOUT ANY WARRANTY; without even the implied warranty of    //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     //
// GNU General Public License for more details.                      //
//                                                                   //
// You should have received a copy of the GNU General Public License //
// along with this program in the file 'gpl.html'; if not, write to  //
// the Free Software Foundation, Inc., 59 Temple Place - Suite 330,  //
// Boston, MA 02111-1307, USA, or contact the author:                //
//                                                                   //
//                       Kevin Kelley  <kelley@iguana.ruralnet.net>  //
//                                                                   //
////////////////////end license & copyright header/////////////////////
package starlight.ui;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.text.*;
import starlight.util.*;

/** Represents a branch of a hierarchical tree.  Can be opened and
  * closed.  The label member is the title of the branch, ie. the
  * name of the directory; and is not itself a subbranch.  It is always
  * visible.
  */
public class Branch extends Container
    implements LabelledItem, LayoutManager, MouseListener, MouseMotionListener
{
  protected boolean enabled   = true;
  protected boolean pressed   = false;
  protected boolean mouseover = false;
  String actionCommand;
  Image openImage;
  Image closedImage;

    int leading = 9;  //magic number to control size of clicky-box and indent
    Leaf label;
    boolean isOpen;
    boolean isSorted = false;

    Branch(Branch parent, String label, String imgOpen, String imgClosed)
        {   this(parent, label, images.ImageLoader.loadImage(imgOpen),
                                images.ImageLoader.loadImage(imgClosed));
        }
    Branch(Branch parent, String lbl, Image openImage, Image closedImage) {
        this.openImage = openImage;
        this.closedImage = closedImage;
        isOpen = (parent == null);
        setLayout(this);
        label = new Leaf(this, lbl, (isOpen? openImage: closedImage));
        label.setVisible(true);
        if (parent!=null)
            parent.addChild(this);
        super.addMouseListener(this);       // we need all mouses; everybody
        super.addMouseMotionListener(this); // else gets forwarded to label.
    }
    public String getLabel() { return label.getLabel(); }
    public void setLabel(String s) { label.setLabel(s); }

    public void addChild(Component node)
        { add(node); isSorted = false; node.setVisible(isOpen); }

    public Component getChild(String name) {
        int nmembers = getComponentCount();
        for (int i=1; i<nmembers; i++) {
            Component m = getComponent(i);
            if (m instanceof LabelledItem)
                if (((LabelledItem)m).getLabel().equals(name))
                    return m;
        }
        return null;
    }
    public void removeChild(String name) {
        int nmembers = getComponentCount();
        for (int i=1; i<nmembers; i++) {
            Component m = getComponent(i);
            if ((m instanceof LabelledItem)
            && (((LabelledItem)m).getLabel().equals(name))) {
                remove(m);
                isSorted = false;
                return;
        }   }
        return;
    }
    // count self and children, recursively
    public int countItems() {
        int count = 1;
        int nmembers = getComponentCount();
        for (int i=1; i<nmembers; i++) {
            Component m = getComponent(i);
            if (m instanceof Branch)
                count += ((Branch)m).countItems();
            else ++count;
        } return count;
    }
    public Component getSubComponent(int index) {
        Stack stack = new Stack();
        stack.push(this);
        int ix=0;
        Component top = (Component)stack.peek();
        while ((top instanceof Branch)
            || (ix < index && !stack.empty())) {
            if (top instanceof Branch) {
                Branch br = (Branch) stack.pop();
                for (int i=br.getComponentCount()-1; i>=0; i--)
                    stack.push(br.getComponent(i));
            } else {
                ++ix; stack.pop();
            }
            top = (Component)stack.peek();
        }
        if (stack.empty()) return null;
        return (Component) stack.pop();  // TOS is desired elt
    }
    public Branch getRoot()
        {   Branch node = this;
            while (node.getParent() instanceof Branch)
                node = (Branch) node.getParent();
            return node;
        }
    public void toggleOpen() {
        isOpen = !isOpen;
        label.setImage(isOpen? openImage: closedImage);

        int nmembers = getComponentCount();
        for (int i=1; i<nmembers; i++)
            getComponent(i).setVisible(isOpen);
        invalidate();
        revalidate();
    }
    void revalidate() {
        Component c = this;
        while (c.getParent()!=null)
            c = c.getParent();
        c.validate();
    }
    protected void validateTree() {
        if (isOpen && !isSorted) {
            sortChildren();
            isSorted = true;
        } super.validateTree();
    }
    boolean boxActive() { return enabled &&             mouseover; }
    boolean boxPushed() { return enabled &&  pressed && mouseover; }

    public void setActionCommand(String s) { actionCommand = s; }
    public String getActionCommand() { return actionCommand; }

    public void  makeEnabled() { enabled = true;  repaint(); }
    public void makeDisabled() { enabled = false; repaint(); }


    public void update(Graphics g) { paint(g); }
    public void  paint(Graphics g) {
        drawDescender(g);
        super.paint(g);
    }
    void drawDescender(Graphics g) {
        Color bg = getBackground();
        boolean isLight = starlight.Colors.intensity(bg) > 0.6;
        Color textColor = isLight? Color.black: Color.white;
        Color butnColor = isLight? Color.white: Color.black;
        Color lineColor = null;
        if (isLight)
              lineColor = boxPushed()? Color.black
                        : boxActive()? bg.darker().darker()
                        : bg.darker();
        else  lineColor = boxPushed()? Color.white
                        : boxActive()? bg.brighter().brighter()
                        : bg.brighter();

        Dimension lblDim = label.getPreferredSize();
        int boxW = leading;
        int boxH = boxW;
        int boxX = 0;
        int boxY = 3*lblDim.height/4 -boxH;
        g.setColor(butnColor);
        g.fillRect(boxX, boxY, boxW, boxH);
        g.setColor(lineColor);
        g.drawRect(boxX, boxY, boxW-1, boxH-1);

        g.setColor(textColor);
        g.drawLine(boxX+2, boxY+boxH/2, boxX+boxW-3, boxY+boxH/2);  // dash
        if (!isOpen)
         g.drawLine(boxX+boxW/2, boxY+2, boxX+boxW/2, boxY+boxH-3);  // bar

        int lineX = boxX + boxW/2;
        int lineY = boxY + boxH;
        g.setColor(lineColor);

        int nmembers = getComponentCount();
        for (int i=1; i<nmembers; i++) {
            Component m = getComponent(i);
            if (m.isVisible()) {
                Rectangle r = m.getBounds();
                int y = r.y;
                if (m instanceof Branch)
                     y += ((Branch)m).getComponent(0).getBounds().height/2-1;
                else y += r.height/2-1;
                g.drawLine(lineX, lineY, lineX, y);
                g.drawLine(lineX, y, r.x, y);
        }   }
    }
    void drawDescender() {
        Graphics g = getGraphics();
        if (g != null) try {
            drawDescender(g);
        } finally { g.dispose(); }
    }
    //---------- MouseListener Implementation ------------------
    public void mouseClicked(MouseEvent e) {}

    public void mouseEntered(MouseEvent e)
        { mouseover = inActiveArea(e.getX(), e.getY()); drawDescender(); }

    public void mouseExited(MouseEvent e)
        { mouseover = false; drawDescender(); }

    public void mousePressed(MouseEvent e)
        {   if (enabled && mouseover) {
                pressed = true;
                toggleOpen();
                repaint();
        }   }
    public void mouseReleased(MouseEvent e)
        {   pressed = false;
            drawDescender();
        }
    //---------- MouseMotionListener Implementation ------------------
    public void mouseDragged(MouseEvent e)
        {   boolean over = inActiveArea(e.getX(), e.getY());
            if (mouseover != over) {
                mouseover = over;
                drawDescender();
        }   }
    public void mouseMoved(MouseEvent e)
        {   boolean over = inActiveArea(e.getX(), e.getY());
            if (mouseover != over) {
                mouseover = over;
                drawDescender();
        }   }
    boolean inActiveArea(int x, int y)
        { return x>=0 && x<leading; }

    //--------------- LayoutManager Implementation ------------//
    public void addLayoutComponent(String name, Component comp) {}
    public void removeLayoutComponent(Component comp) {}
    public Dimension minimumLayoutSize(Container target)
        { return preferredLayoutSize(target); }
    public Dimension preferredLayoutSize(Container target) {
        Dimension dim = new Dimension(leading,0);
        int nmembers = target.getComponentCount();
        for (int i=0; i<nmembers; i++) {
            Component m = target.getComponent(i);
            if (m.isVisible()) {
                int xoffset = (i==0)? leading: 2*leading;
                Dimension d = m.getPreferredSize();
                dim.width = Math.max(dim.width, d.width+xoffset);
                dim.height += d.height;
        }   }
        return dim;
    }
    public void layoutContainer(Container target) {
        int yoffset = 0;
        int nmembers = target.getComponentCount();
        for (int i=0; i<nmembers; i++) {
            Component m = target.getComponent(i);
            if (m.isVisible()) {
                int xoffset = (i==0)? leading: 2*leading;
                Dimension d = m.getPreferredSize();
                m.setBounds(xoffset, yoffset, d.width, d.height);
                yoffset += d.height;
        }   }
        repaint();
    }
    public void addActionListener(ActionListener listener)
        { label.addActionListener(listener); }
    public void removeActionListener(ActionListener listener)
        { label.removeActionListener(listener); }

    public void addMouseListener(MouseListener listener)
        { label.addMouseListener(listener); }
    public void removeMouseListener(MouseListener listener)
        { label.removeMouseListener(listener); }

    public void addKeyListener(KeyListener listener)
        { label.addKeyListener(listener); }
    public void removeKeyListener(KeyListener listener)
        { label.removeKeyListener(listener); }


private void sortChildren() {
    // remove all children (not the label) and stash them in a table
    int nchildren = getComponentCount()-1;
    Hashtable children = new Hashtable();
    String[] childnames = new String[nchildren];
    for (int i=nchildren-1; i>=0; --i) {
        int index = i+1;
        LabelledItem item = (LabelledItem)getComponent(index);
        String itemname = item.getLabel();
        childnames[i] = itemname;
        children.put(itemname, item);
        remove(index);
    }
    // generate sort keys for all child names
    Collator collator = Collator.getInstance();
    CollationKey[] keys = new CollationKey[nchildren];
    for (int i=0; i<nchildren; i++)
        keys[i] = collator.getCollationKey(childnames[i]);

    // sort 'em
    ArraySorter.sort(keys, new Comparator() {
        public boolean less(Object lhs, Object rhs)
        { return (((CollationKey)lhs).compareTo((CollationKey)rhs) < 0); }
    });
    // add children back, in order
    int marker = 0;
    for (int i=0; i<keys.length; i++) {         // pull out the dirs first...
        String name = keys[i].getSourceString();
        LabelledItem item = (LabelledItem)children.get(name);
        if (item instanceof Branch) {
            add((Component)item);
            keys[i] = null;
    }   }
    for (int i=0; i<keys.length; i++) {        // what's left are the non-dirs.
        if (keys[i] != null) {
            String name = keys[i].getSourceString();
            LabelledItem item = (LabelledItem)children.get(name);
            add((Component)item);
    }   }
}
}
