//////////////////////license & copyright header///////////////////////
//                                                                   //
//              Java Cannery - jar management utility.               //
//                Copyright (c) 1998 by Kevin Kelley                 //
//                                                                   //
// This program is free software; you can redistribute it and/or     //
// modify it under the terms of the GNU General Public License as    //
// published by the Free Software Foundation; either version 2 of    //
// the License, or (at your option) any later version.               //
//                                                                   //
// This program is distributed in the hope that it will be useful,   //
// but WITHOUT ANY WARRANTY; without even the implied warranty of    //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     //
// GNU General Public License for more details.                      //
//                                                                   //
// You should have received a copy of the GNU General Public License //
// along with this program in the file 'gpl.html'; if not, write to  //
// the Free Software Foundation, Inc., 59 Temple Place - Suite 330,  //
// Boston, MA 02111-1307, USA, or contact the author:                //
//                                                                   //
//                       Kevin Kelley  <kelley@iguana.ruralnet.net>  //
//                                                                   //
////////////////////end license & copyright header/////////////////////
package starlight.util;
import java.io.InputStream;
import java.io.IOException;


public class ByteArray {

  public static byte[] readFully(InputStream in) throws IOException {
    ByteArray ba = new ByteArray();
    ba.appendFrom(in);
    if (ba.array.length == ba.count)  // avoid copy if possible
        return ba.getArrayRep();
    return ba.toByteArray();
  }

    byte[]      array;
    int         count;

    /** Constructs a array with an initial capacity of 8 elements.
      */
    public ByteArray() {
        this(8);
    }

    /** Primitive constructor. Constructs an array large enough
      * to hold <b>initialCapacity</b> elements.
      */
    public ByteArray(int initialCapacity) {
        array = new byte[initialCapacity];
    }

    /** Returns the number of bytes in the array.
      */
    public int size() {
        return count;
    }

    /** Adjust size for when array is manipulated directly.
      */
    public void setSize(int n) {
       ensureCapacity(n);
       count = n;
    }

    /** Returns <b>true</b> if the array contains no elements.
      */
    public boolean isEmpty() {
        return (count == 0);
    }

    /** Adds elements contained in a to end of array.
      */
    public void appendFrom(byte[] a) {
        int addCount = a.length;

        if (array.length < count + addCount)
            ensureCapacity(count + addCount);

        System.arraycopy(a, 0, array, count, addCount);

        count += addCount;
    }

    /** Adds all the bytes from an input stream.  (available() may lie,
      * so we'll keep asking it, but we'll trust the answer enough to
      * ensure buffer size.
      */
    public void appendFrom(InputStream is) throws IOException {
        int avail;
        while ((avail=is.available()) > 0) {
            if (array.length < count + avail)
                ensureCapacity(count + avail);
            int read = is.read(array, count, avail);
            if (read <= 0)  // 0 if read all available; -1 if hit eos
                break;
            count += read;
        }
    }

    /** Adds <b>numbytes</b> bytes from an input stream.  Loop until the
      * correct number of bytes have been read, or EOS reached.
      */
    public void appendFrom(InputStream is, int numbytes) throws IOException {
      if (array.length < count + numbytes)
         ensureCapacity (count + numbytes);
       while (numbytes > 0) {
          int read = is.read(array, count, numbytes);
          if (read == -1)
              break;
          count += read;
          numbytes -= read;
       }
    }

    /** Returns a properly sized copy of the array's contents.
      */
    public byte[] toByteArray() {
        byte[] newArray = new byte[count];
        System.arraycopy(array, 0, newArray, 0, count);
        return newArray;
    }

    /** Returns the actual array containing the array's contents.
      */
    public byte[] getArrayRep() {
        return array;
    }

    /** Increases, if necessary, the the array's storage area so that it
      * can contain <b>minCapacity</b> elements.
      */
    public void ensureCapacity(int minCapacity) {
        if (array.length < minCapacity) {
            byte[] newArray = new byte[minCapacity];
            System.arraycopy(array, 0, newArray, 0, count);
            array = newArray;
        }
    }

    /** Decreases the storage occupied by this array to the minimum.
      */
    public void trimToSize() {
        if (array.length > count) {
            byte[] newArray = new byte[count];
            System.arraycopy(array, 0, newArray, 0, count);
            array = newArray;
        }
    }

    /** Returns the number of elements that can be stored in the array
      * without increasing the array's storage area.
      */
    public int capacity() {
        return array.length;
    }

    /** Empties the array, but leaves its capacity unchanged.
      */
    public void removeAllElements() {
        count = 0;
    }

    /** Returns the array's string representation.
      */
    public String toString() {
        return new String(array, 0, count);
    }
}
